@@ -135,8 +135,6 @@ struct obj_zwp_linux_dmabuf_params {
 		uint32_t offset;
 		uint32_t stride;
 		uint64_t modifier;
-		char *msg;
-		int msg_len;
 	} add[MAX_DMABUF_PLANES];
 	int nplanes;
 };
@@ -278,9 +276,6 @@ void destroy_wp_object(struct wp_object *object)
 					}
 				}
 			}
-			if (r->add[i].msg) {
-				free(r->add[i].msg);
-			}
 		}
 	} else if (object->type == &intf_zwlr_export_dmabuf_frame_v1) {
 		struct obj_wlr_export_dmabuf_frame *r =
@@ -486,7 +481,7 @@ void do_wl_registry_req_bind(struct context *ctx, uint32_t name,
 	}
 
 fail:
-	wp_debug("Binding fail name=%d %s id=%d (v%d)", name, interface,
+	wp_debug("Unhandled protocol %s name=%d id=%d (v%d)", interface, name,
 			the_object->obj_id, version);
 
 	tracker_remove(ctx->tracker, the_object);
@@ -497,7 +492,7 @@ fail:
 }
 
 void do_wl_buffer_evt_release(struct context *ctx) { (void)ctx; }
-static int get_shm_bytes_per_pixel(uint32_t format)
+int get_shm_bytes_per_pixel(uint32_t format)
 {
 	switch (format) {
 	case 0x34325241: /* DRM_FORMAT_ARGB8888 */
@@ -1244,6 +1239,42 @@ void do_wl_drm_req_create_prime_buffer(struct context *ctx,
 	buf->dmabuf_offsets[0] = (uint32_t)offset0;
 	buf->dmabuf_strides[0] = (uint32_t)stride0;
 	buf->unique_id = ctx->g->tracker.buffer_seqno++;
+
+	if (ctx->on_display_side) {
+		/* the new dmabuf being created is not guaranteed to
+		 * have the original offset/stride parameters, so reset
+		 * them */
+		ctx->message[6] = 0;
+		ctx->message[7] = dmabuf_get_stride(sfd->dmabuf_bo);
+	}
+}
+
+static bool dmabuf_format_permitted(
+		struct context *ctx, uint32_t format, uint64_t modifier)
+{
+	if (ctx->g->config->only_linear_dmabuf) {
+		/* MOD_INVALID is allowed because some drivers don't support
+		 * LINEAR. Every modern GPU+driver should be able to handle
+		 * LINEAR. Conditionally blocking INVALID (i.e, if LINEAR is an
+		 * option) can break things when the application-side Waypipe
+		 * instance does not support LINEAR. */
+		if (modifier != 0 && modifier != DRM_FORMAT_MOD_INVALID) {
+			return false;
+		}
+	}
+	/* Filter out formats which are not recognized, or multiplane */
+	if (get_shm_bytes_per_pixel(format) == -1) {
+		return false;
+	}
+	/* Blacklist intel modifiers which introduce a second color control
+	 * surface; todo: add support for these, eventually */
+	if (modifier == (1uLL << 56 | 4) || modifier == (1uLL << 56 | 5) ||
+			modifier == (1uLL << 56 | 6) ||
+			modifier == (1uLL << 56 | 7) ||
+			modifier == (1uLL << 56 | 8)) {
+		return false;
+	}
+	return true;
 }
 
 void do_zwp_linux_dmabuf_v1_evt_modifier(struct context *ctx, uint32_t format,
@@ -1253,13 +1284,7 @@ void do_zwp_linux_dmabuf_v1_evt_modifier(struct context *ctx, uint32_t format,
 	(void)format;
 	uint64_t modifier = modifier_hi * 0x100000000uLL + modifier_lo;
 	// Prevent all advertisements for dmabufs with modifiers
-	if (ctx->g->config->only_linear_dmabuf) {
-		if (modifier != 0 && modifier != DRM_FORMAT_MOD_INVALID) {
-			ctx->drop_this_msg = true;
-		}
-	}
-	/* Filter out formats which are not recognized, or multiplane */
-	if (get_shm_bytes_per_pixel(format) == -1) {
+	if (!dmabuf_format_permitted(ctx, format, modifier)) {
 		ctx->drop_this_msg = true;
 	}
 }
@@ -1325,102 +1350,33 @@ void do_zwp_linux_buffer_params_v1_req_add(struct context *ctx, int fd,
 	params->add[plane_idx].modifier =
 			modifier_lo + modifier_hi * 0x100000000uLL;
 	// Only perform rearrangement on the client side, for now
-	if (!ctx->on_display_side) {
-		params->add[plane_idx].msg =
-				malloc((size_t)ctx->message_length);
-		memcpy(params->add[plane_idx].msg, ctx->message,
-				(size_t)ctx->message_length);
-		params->add[plane_idx].msg_len = ctx->message_length;
-
+	if (true) {
 		ctx->drop_this_msg = true;
 	}
 }
-static int reintroduce_add_msgs(struct context *context,
-		struct obj_zwp_linux_dmabuf_params *params)
-{
-	int net_length = context->message_length;
-	int nfds = 0;
-	for (int i = 0; i < params->nplanes; i++) {
-		net_length += params->add[i].msg_len;
-		nfds++;
-	}
-	if (net_length > context->message_available_space) {
-		wp_error("Not enough space to reintroduce zwp_linux_buffer_params_v1.add message data");
-		return -1;
-	}
-	if (nfds > context->fds->size - context->fds->zone_end) {
-		wp_error("Not enough space to reintroduce zwp_linux_buffer_params_v1.add message fds");
-		return -1;
-	}
-	// Update fds
-	int nmoved = (context->fds->zone_end - context->fds->zone_start);
-	memmove(context->fds->data + context->fds->zone_start + nfds,
-			context->fds->data + context->fds->zone_start,
-			(size_t)nmoved * sizeof(int));
-	for (int i = 0; i < params->nplanes; i++) {
-		context->fds->data[context->fds->zone_start + i] =
-				params->add[i].fd;
-	}
-	/* We inject `nfds` new file descriptors, and advance the zone
-	 * of queued file descriptors forward, since the injected file
-	 * descriptors will not be used by the parser, but will still
-	 * be transported out. */
-	context->fds->zone_start += nfds;
-	context->fds->zone_end += nfds;
 
-	// Update data
-	char *cmsg = (char *)context->message;
-	memmove(cmsg + net_length - context->message_length, cmsg,
-			(size_t)context->message_length);
-	int start = 0;
-	for (int i = 0; i < params->nplanes; i++) {
-		memcpy(cmsg + start, params->add[i].msg,
-				(size_t)params->add[i].msg_len);
-		/* Tag the message as having one file descriptor */
-		((uint32_t *)(cmsg + start))[1] |= (uint32_t)(1 << 11);
-		start += params->add[i].msg_len;
-		free(params->add[i].msg);
-		params->add[i].msg = NULL;
-		params->add[i].msg_len = 0;
-	}
-	wp_debug("Reintroducing add requests for zwp_linux_buffer_params_v1, going from %d to %d bytes",
-			context->message_length, net_length);
-	context->message_length = net_length;
-	context->fds_changed = true;
-	return 0;
-}
-/** After this function is called, all subsets of fds that duplicate an
- * underlying dmabuf will be reduced to select a single fd. */
-static void deduplicate_dmabuf_fds(struct context *context,
-		struct obj_zwp_linux_dmabuf_params *params)
+static uint32_t append_zwp_linux_buffer_params_v1_req_add(uint32_t *msg,
+		bool display_side, uint32_t obj_id, uint32_t plane_idx,
+		uint32_t offset, uint32_t stride, uint32_t modifier_hi,
+		uint32_t modifier_lo)
 {
-	if (params->nplanes == 1) {
-		return;
-	}
-	int handles[MAX_DMABUF_PLANES];
-	struct gbm_bo *temp_bos[MAX_DMABUF_PLANES];
-	memset(temp_bos, 0, sizeof(temp_bos));
-	for (int i = 0; i < params->nplanes; i++) {
-		handles[i] = get_unique_dmabuf_handle(&context->g->render,
-				params->add[i].fd, &temp_bos[i]);
-	}
-	for (int i = 0; i < params->nplanes; i++) {
-		destroy_dmabuf(temp_bos[i]);
-	}
-	for (int i = 0; i < params->nplanes; i++) {
-		int lowest = i;
-		for (int k = 0; k < i; k++) {
-			if (handles[k] == handles[i]) {
-				lowest = k;
-				break;
-			}
+	uint32_t msg_size = 2;
+	if (msg) {
+		msg[0] = obj_id;
+		msg[msg_size++] = plane_idx;
+		msg[msg_size++] = offset;
+		msg[msg_size++] = stride;
+		msg[msg_size++] = modifier_hi;
+		msg[msg_size++] = modifier_lo;
+		msg[1] = ((uint32_t)msg_size << 18) | 1;
+		/* Tag the message as having one file descriptor */
+		if (!display_side) {
+			msg[1] |= (uint32_t)(1 << 11);
 		}
-		if (lowest != i &&
-				params->add[i].fd != params->add[lowest].fd) {
-			checked_close(params->add[i].fd);
-		}
-		params->add[i].fd = params->add[lowest].fd;
+	} else {
+		msg_size += 5;
 	}
+	return msg_size;
 }
 
 void do_zwp_linux_buffer_params_v1_req_create(struct context *ctx,
@@ -1432,10 +1388,7 @@ void do_zwp_linux_buffer_params_v1_req_create(struct context *ctx,
 	params->create_width = width;
 	params->create_height = height;
 	params->create_format = format;
-	deduplicate_dmabuf_fds(ctx, params);
-	if (!ctx->on_display_side) {
-		reintroduce_add_msgs(ctx, params);
-	}
+
 	struct dmabuf_slice_data info = {.width = (uint32_t)width,
 			.height = (uint32_t)height,
 			.format = format,
@@ -1474,12 +1427,23 @@ void do_zwp_linux_buffer_params_v1_req_create(struct context *ctx,
 			}
 		}
 
+		/* note: the``info` provided includes the incoming/as-if stride
+		 * data. */
 		struct shadow_fd *sfd = translate_fd(&ctx->g->map,
 				&ctx->g->render, params->add[i].fd, res_type, 0,
 				&info, false);
 		if (!sfd) {
 			continue;
 		}
+		if (ctx->on_display_side) {
+			/* the new dmabuf being created is not guaranteed to
+			 * have the original offset/stride parameters, so reset
+			 * them */
+			params->add[i].offset = 0;
+			params->add[i].stride =
+					dmabuf_get_stride(sfd->dmabuf_bo);
+		}
+
 		/* increment for each extra time this fd will be sent */
 		if (sfd->has_owner) {
 			shadow_incref_transfer(sfd);
@@ -1487,6 +1451,69 @@ void do_zwp_linux_buffer_params_v1_req_create(struct context *ctx,
 		// Convert the stored fds to buffer pointers now.
 		params->add[i].buffer = shadow_incref_protocol(sfd);
 	}
+
+	if (true) {
+		// Update file descriptors
+		int nfds = params->nplanes;
+		if (nfds > ctx->fds->size - ctx->fds->zone_end) {
+			wp_error("Not enough space to reintroduce zwp_linux_buffer_params_v1.add message fds");
+			return;
+		}
+		int nmoved = (ctx->fds->zone_end - ctx->fds->zone_start);
+		memmove(ctx->fds->data + ctx->fds->zone_start + nfds,
+				ctx->fds->data + ctx->fds->zone_start,
+				(size_t)nmoved * sizeof(int));
+		for (int i = 0; i < params->nplanes; i++) {
+			ctx->fds->data[ctx->fds->zone_start + i] =
+					params->add[i].fd;
+		}
+		/* We inject `nfds` new file descriptors, and advance the zone
+		 * of queued file descriptors forward, since the injected file
+		 * descriptors will not be used by the parser, but will still
+		 * be transported out. */
+		ctx->fds->zone_start += nfds;
+		ctx->fds->zone_end += nfds;
+		ctx->fds_changed = true;
+
+		// Update data
+		int net_length = ctx->message_length;
+		uint32_t extra = 0;
+		for (int i = 0; i < params->nplanes; i++) {
+			extra += append_zwp_linux_buffer_params_v1_req_add(NULL,
+					ctx->on_display_side,
+					params->base.obj_id, (uint32_t)i,
+					params->add[i].offset,
+					params->add[i].stride,
+					(uint32_t)(params->add[i].modifier >>
+							32),
+					(uint32_t)(params->add[i].modifier));
+		}
+		net_length += (int)(sizeof(uint32_t) * extra);
+		if (net_length > ctx->message_available_space) {
+			wp_error("Not enough space to reintroduce zwp_linux_buffer_params_v1.add message data");
+			return;
+		}
+		char *cmsg = (char *)ctx->message;
+		memmove(cmsg + net_length - ctx->message_length, cmsg,
+				(size_t)ctx->message_length);
+		size_t start = 0;
+		for (int i = 0; i < params->nplanes; i++) {
+			uint32_t step = append_zwp_linux_buffer_params_v1_req_add(
+					(uint32_t *)(cmsg + start),
+					ctx->on_display_side,
+					params->base.obj_id, (uint32_t)i,
+					params->add[i].offset,
+					params->add[i].stride,
+					(uint32_t)(params->add[i].modifier >>
+							32),
+					(uint32_t)(params->add[i].modifier));
+			start += step * sizeof(uint32_t);
+		}
+		wp_debug("Reintroducing add requests for zwp_linux_buffer_params_v1, going from %d to %d bytes",
+				ctx->message_length, net_length);
+		ctx->message_length = net_length;
+	}
+
 	// Avoid closing in destroy_wp_object
 	for (int i = 0; i < MAX_DMABUF_PLANES; i++) {
 		params->add[i].fd = -1;
@@ -1509,7 +1536,6 @@ void do_zwp_linux_dmabuf_feedback_v1_evt_done(struct context *ctx)
 	struct obj_zwp_linux_dmabuf_feedback *obj =
 			(struct obj_zwp_linux_dmabuf_feedback *)ctx->obj;
 
-	bool has_any_linear = false;
 	int worst_case_space = 2;
 	for (size_t i = 0; i < obj->tranche_count; i++) {
 		for (size_t j = 0; j < obj->tranches[i].tranche_size; j++) {
@@ -1519,10 +1545,6 @@ void do_zwp_linux_dmabuf_feedback_v1_evt_done(struct context *ctx)
 						idx, obj->table_len);
 				return;
 			}
-			uint64_t modifier = obj->table[idx].modifier;
-			if (modifier == 0) {
-				has_any_linear = true;
-			}
 		}
 
 		worst_case_space +=
@@ -1544,17 +1566,8 @@ void do_zwp_linux_dmabuf_feedback_v1_evt_done(struct context *ctx)
 		uint16_t *fmts = (uint16_t *)&ctx->message[m + 3];
 		for (size_t j = 0; j < obj->tranches[i].tranche_size; j++) {
 			uint16_t idx = obj->tranches[i].tranche[j];
-			uint64_t modifier = obj->table[idx].modifier;
-
-			bool is_single_plane =
-					get_shm_bytes_per_pixel(
-							obj->table[idx].format) >
-					0;
-			bool keep = is_single_plane &&
-				    (modifier == 0 ||
-						    (!has_any_linear &&
-								    modifier == DRM_FORMAT_MOD_INVALID));
-			if (keep) {
+			if (dmabuf_format_permitted(ctx, obj->table[idx].format,
+					    obj->table[idx].modifier)) {
 				fmts[w++] = idx;
 			}
 		}
