@@ -91,6 +91,10 @@ static void destroy_unlinked_sfd(struct shadow_fd *sfd)
 		}
 		destroy_dmabuf(sfd->dmabuf_bo);
 		zeroed_aligned_free(sfd->mem_mirror, &sfd->mem_mirror_handle);
+		if (sfd->dmabuf_warped_handle) {
+			zeroed_aligned_free(sfd->dmabuf_warped,
+					&sfd->dmabuf_warped_handle);
+		}
 	} else if (sfd->type == FDC_PIPE) {
 		if (sfd->pipe.fd != sfd->fd_local && sfd->pipe.fd != -1) {
 			checked_close(sfd->pipe.fd);
@@ -708,7 +712,7 @@ static void *shrink_buffer(void *buf, size_t sz)
 }
 
 /* Construct and optionally compress a diff between sfd->mem_mirror and
- * the actual memmap'd data */
+ * the actual memmap'd data, and synchronize sfd->mem_mirror */
 static void worker_run_compress_diff(
 		struct task_data *task, struct thread_data *local)
 {
@@ -747,15 +751,61 @@ static void worker_run_compress_diff(
 	}
 
 	DTRACE_PROBE1(waypipe, construct_diff_enter, task->damage_len);
+	char *source = sfd->mem_local;
+	if (sfd->type == FDC_DMABUF &&
+			sfd->dmabuf_map_stride != sfd->dmabuf_info.strides[0]) {
+		size_t tx_stride = (size_t)sfd->dmabuf_info.strides[0];
+		size_t common = (size_t)minu(sfd->dmabuf_map_stride, tx_stride);
+		/* copy mapped data to temporary buffer whose stride matches
+		 * what is sent over the wire */
+		for (int i = 0; i < task->damage_len; i++) {
+			size_t start = (size_t)task->damage_intervals[i].start;
+			size_t end = (size_t)task->damage_intervals[i].end;
+			size_t loc_start =
+					(start % tx_stride) +
+					(start / tx_stride) *
+							sfd->dmabuf_map_stride;
+			size_t loc_end = (end % tx_stride) +
+					 (end / tx_stride) *
+							 sfd->dmabuf_map_stride;
+
+			stride_shifted_copy(sfd->dmabuf_warped, sfd->mem_local,
+					loc_start, loc_end - loc_start, common,
+					sfd->dmabuf_map_stride,
+					sfd->dmabuf_info.strides[0]);
+		}
+
+		if (task->damaged_end) {
+			size_t alignment = 1u << pool->diff_alignment_bits;
+			size_t start = alignment *
+				       (sfd->buffer_size / alignment);
+			size_t end = sfd->buffer_size;
+
+			size_t loc_start =
+					(start % tx_stride) +
+					(start / tx_stride) *
+							sfd->dmabuf_map_stride;
+			size_t loc_end = (end % tx_stride) +
+					 (end / tx_stride) *
+							 sfd->dmabuf_map_stride;
+
+			stride_shifted_copy(sfd->dmabuf_warped, sfd->mem_local,
+					loc_start, loc_end - loc_start, common,
+					sfd->dmabuf_map_stride,
+					sfd->dmabuf_info.strides[0]);
+		}
+
+		source = sfd->dmabuf_warped;
+	}
+
 	size_t diffsize = construct_diff_core(pool->diff_func,
 			pool->diff_alignment_bits, task->damage_intervals,
-			task->damage_len, sfd->mem_mirror, sfd->mem_local,
-			diff_target);
+			task->damage_len, sfd->mem_mirror, source, diff_target);
 	size_t ntrailing = 0;
 	if (task->damaged_end) {
 		ntrailing = construct_diff_trailing(sfd->buffer_size,
 				pool->diff_alignment_bits, sfd->mem_mirror,
-				sfd->mem_local, diff_target + diffsize);
+				source, diff_target + diffsize);
 	}
 	DTRACE_PROBE1(waypipe, construct_diff_exit, diffsize);
 
@@ -801,11 +851,10 @@ end:
 	DTRACE_PROBE1(waypipe, worker_compdiff_exit, diffsize);
 }
 
-/* Compress data for sfd->mem_mirror */
+/* Compress data for sfd->mem_mirror, and synchronize sfd->mem_mirror */
 static void worker_run_compress_block(
 		struct task_data *task, struct thread_data *local)
 {
-
 	struct shadow_fd *sfd = task->sfd;
 	struct thread_pool *pool = local->pool;
 	if (task->zone_end == task->zone_start) {
@@ -818,6 +867,29 @@ static void worker_run_compress_block(
 	size_t source_end = (size_t)task->zone_end;
 	DTRACE_PROBE1(waypipe, worker_comp_enter, source_end - source_start);
 
+	/* Update mirror to match local */
+	if (sfd->type == FDC_DMABUF &&
+			sfd->dmabuf_map_stride != sfd->dmabuf_info.strides[0]) {
+		uint32_t tx_stride = sfd->dmabuf_info.strides[0];
+		size_t common = (size_t)minu(sfd->dmabuf_map_stride,
+				sfd->dmabuf_info.strides[0]);
+		size_t loc_start = (source_start % tx_stride) +
+				   (source_start / tx_stride) *
+						   sfd->dmabuf_map_stride;
+		size_t loc_end = (source_end % tx_stride) +
+				 (source_end / tx_stride) *
+						 sfd->dmabuf_map_stride;
+
+		stride_shifted_copy(sfd->mem_mirror, sfd->mem_local, loc_start,
+				loc_end - loc_start, common,
+				sfd->dmabuf_map_stride,
+				sfd->dmabuf_info.strides[0]);
+	} else {
+		memcpy(sfd->mem_mirror + source_start,
+				sfd->mem_local + source_start,
+				source_end - source_start);
+	}
+
 	size_t sz = 0;
 	uint8_t *msg;
 	if (pool->compression == COMP_NONE) {
@@ -1182,7 +1254,8 @@ void collect_update(struct thread_pool *threads, struct shadow_fd *sfd,
 		}
 		if (!sfd->mem_local) {
 			sfd->mem_local = map_dmabuf(sfd->dmabuf_bo, false,
-					&sfd->dmabuf_map_handle, NULL, NULL);
+					&sfd->dmabuf_map_handle,
+					&sfd->dmabuf_map_stride);
 			if (!sfd->mem_local) {
 				return;
 			}
@@ -1192,12 +1265,13 @@ void collect_update(struct thread_pool *threads, struct shadow_fd *sfd,
 			sfd->mem_mirror = zeroed_aligned_alloc(
 					alignz(sfd->buffer_size, alignment),
 					alignment, &sfd->mem_mirror_handle);
-			if (!sfd->mem_mirror) {
+			sfd->dmabuf_warped = zeroed_aligned_alloc(
+					alignz(sfd->buffer_size, alignment),
+					alignment, &sfd->dmabuf_warped_handle);
+			if (!sfd->mem_mirror || !sfd->dmabuf_warped) {
 				wp_error("Failed to allocate mirror");
 				return;
 			}
-			memcpy(sfd->mem_mirror, sfd->mem_local,
-					sfd->buffer_size);
 
 			sfd->remote_bufsize = 0;
 			queue_fill_transfers(threads, sfd, transfers);
@@ -1205,7 +1279,6 @@ void collect_update(struct thread_pool *threads, struct shadow_fd *sfd,
 		} else {
 			// TODO: detailed damage tracking
 			damage_everything(&sfd->damage);
-
 			queue_diff_transfers(threads, sfd, transfers);
 		}
 		/* Unmapping will be handled by finish_update() */
@@ -1546,18 +1619,23 @@ int apply_update(struct fd_translation_map *map, struct thread_pool *threads,
 		}
 
 		sfd->type = FDC_DMABUF;
-		const struct wmsg_open_dmabuf header =
-				*(const struct wmsg_open_dmabuf *)msg->data;
-		sfd->buffer_size = header.file_size;
 
 		memcpy(&sfd->dmabuf_info,
 				msg->data + sizeof(struct wmsg_open_dmabuf),
 				sizeof(struct dmabuf_slice_data));
+		/* allocate a mirror buffer that matches dimensions of incoming
+		 * data from the remote; this may disagree with the mapped size
+		 * of the buffer */
+		sfd->buffer_size = sfd->dmabuf_info.height *
+				   sfd->dmabuf_info.strides[0];
 		size_t alignment = 1u << threads->diff_alignment_bits;
 		sfd->mem_mirror = zeroed_aligned_alloc(
 				alignz(sfd->buffer_size, alignment), alignment,
 				&sfd->mem_mirror_handle);
-		if (!sfd->mem_mirror) {
+		sfd->dmabuf_warped = zeroed_aligned_alloc(
+				alignz(sfd->buffer_size, alignment), alignment,
+				&sfd->dmabuf_warped_handle);
+		if (!sfd->mem_mirror || !sfd->dmabuf_warped) {
 			wp_error("Failed to allocate mirror");
 			return 0;
 		}
@@ -1849,40 +1927,58 @@ int apply_update(struct fd_translation_map *map, struct thread_pool *threads,
 					act_size, header->end - header->start);
 			return ERR_FATAL;
 		}
-		memcpy(sfd->mem_mirror + header->start, act_buffer,
-				header->end - header->start);
 
-		void *handle = NULL;
-		bool already_mapped = sfd->mem_local != NULL;
-		if (sfd->type == FDC_DMABUF && !already_mapped) {
-			uint32_t stride = 0, height = 0;
-			sfd->mem_local = map_dmabuf(sfd->dmabuf_bo, true,
-					&handle, &stride, &height);
-			if (stride * height < sfd->buffer_size) {
-				wp_error("DMABUF mapped with stride %" PRIu32
-					 " height %" PRIu32
-					 ", but expected ize=%zu>%zu, ignoring overlarge update",
-						stride, height,
-						sfd->buffer_size,
-						(size_t)(stride * height));
-				unmap_dmabuf(sfd->dmabuf_bo, handle);
+		if (sfd->type == FDC_DMABUF) {
+			int bpp = get_shm_bytes_per_pixel(
+					sfd->dmabuf_info.format);
+			if (bpp == -1) {
+				wp_error("Skipping update of RID=%d, non-RGBA/monoplane fmt %x",
+						sfd->remote_id,
+						sfd->dmabuf_info.format);
 				return 0;
 			}
-		}
-		if (!sfd->mem_local) {
-			wp_error("Failed to fill RID=%d, fd not mapped",
-					sfd->remote_id);
-			return 0;
-		}
-		memcpy(sfd->mem_local + header->start,
-				sfd->mem_mirror + header->start,
-				header->end - header->start);
 
-		if (sfd->type == FDC_DMABUF && !already_mapped) {
-			sfd->mem_local = NULL;
+			memcpy(sfd->mem_mirror + header->start, act_buffer,
+					header->end - header->start);
+
+			void *handle = NULL;
+			uint32_t map_stride = 0;
+			char *mem_local = map_dmabuf(sfd->dmabuf_bo, true,
+					&handle, &map_stride);
+			if (!mem_local) {
+				wp_error("Failed to apply fill to RID=%d, fd not mapped",
+						sfd->remote_id);
+				return 0;
+			}
+			uint32_t in_stride = sfd->dmabuf_info.strides[0];
+			if (map_stride == in_stride) {
+				memcpy(mem_local + header->start,
+						sfd->mem_mirror + header->start,
+						header->end - header->start);
+			} else {
+				/* stride changing transfer */
+				uint32_t row_length = (uint32_t)bpp *
+						      sfd->dmabuf_info.width;
+
+				uint32_t copy_size = (uint32_t)minu(row_length,
+						minu(map_stride, in_stride));
+
+				stride_shifted_copy(mem_local,
+						act_buffer - header->start,
+						header->start,
+						header->end - header->start,
+						copy_size, in_stride,
+						map_stride);
+			}
+
 			if (unmap_dmabuf(sfd->dmabuf_bo, handle) == -1) {
 				return 0;
 			}
+		} else {
+			memcpy(sfd->mem_mirror + header->start, act_buffer,
+					header->end - header->start);
+			memcpy(sfd->mem_local + header->start, act_buffer,
+					header->end - header->start);
 		}
 		return 0;
 	}
@@ -1928,41 +2024,85 @@ int apply_update(struct fd_translation_map *map, struct thread_pool *threads,
 			return ERR_FATAL;
 		}
 
-		void *handle = NULL;
-		bool already_mapped = sfd->mem_local != NULL;
-		if (sfd->type == FDC_DMABUF && !already_mapped) {
-			uint32_t stride = 0, height = 0;
-			sfd->mem_local = map_dmabuf(sfd->dmabuf_bo, true,
-					&handle, &stride, &height);
-			if (stride * height < sfd->buffer_size) {
-				wp_error("DMABUF mapped with stride %" PRIu32
-					 " height %" PRIu32
-					 ", but expected size=%zu>%zu, ignoring overlarge update",
-						stride, height,
-						sfd->buffer_size,
-						(size_t)(stride * height));
-				unmap_dmabuf(sfd->dmabuf_bo, handle);
+		if (sfd->type == FDC_DMABUF) {
+			int bpp = get_shm_bytes_per_pixel(
+					sfd->dmabuf_info.format);
+			if (bpp == -1) {
+				wp_error("Skipping update of RID=%d, non-RGBA/monoplane fmt %x",
+						sfd->remote_id,
+						sfd->dmabuf_info.format);
 				return 0;
 			}
-		}
-		if (!sfd->mem_local) {
-			wp_error("Failed to apply diff to RID=%d, fd not mapped",
-					sfd->remote_id);
-			return 0;
-		}
 
-		DTRACE_PROBE2(waypipe, apply_diff_enter, sfd->buffer_size,
-				header->diff_size);
-		apply_diff(sfd->buffer_size, sfd->mem_mirror, sfd->mem_local,
-				header->diff_size, header->ntrailing,
-				act_buffer);
-		DTRACE_PROBE(waypipe, apply_diff_exit);
+			void *handle = NULL;
+			uint32_t map_stride = 0;
+			char *mem_local = map_dmabuf(sfd->dmabuf_bo, true,
+					&handle, &map_stride);
+			if (!mem_local) {
+				wp_error("Failed to apply diff to RID=%d, fd not mapped",
+						sfd->remote_id);
+				return 0;
+			}
+			uint32_t in_stride = sfd->dmabuf_info.strides[0];
+			uint32_t row_length =
+					(uint32_t)bpp * sfd->dmabuf_info.width;
+			uint32_t copy_size = (uint32_t)minu(row_length,
+					minu(map_stride, in_stride));
+
+			(void)in_stride;
+			size_t nblocks = sfd->buffer_size / sizeof(uint32_t);
+			size_t ndiffblocks =
+					header->diff_size / sizeof(uint32_t);
+			uint32_t *diff_blocks = (uint32_t *)act_buffer;
+			for (size_t i = 0; i < ndiffblocks;) {
+				size_t nfrom = (size_t)diff_blocks[i];
+				size_t nto = (size_t)diff_blocks[i + 1];
+				size_t span = nto - nfrom;
+				if (nto > nblocks || nfrom >= nto ||
+						i + (nto - nfrom) >=
+								ndiffblocks) {
+					wp_error("Invalid copy range [%zu,%zu) > %zu=nblocks or [%zu,%zu) > %zu=ndiffblocks",
+							nfrom, nto, nblocks,
+							i + 1, i + 1 + span,
+							ndiffblocks);
+					break;
+				}
+				memcpy(sfd->mem_mirror + sizeof(uint32_t) * nfrom,
+						diff_blocks + i + 2,
+						sizeof(uint32_t) * span);
+				stride_shifted_copy(mem_local,
+						(char *)((diff_blocks + i + 2) -
+								nfrom),
+						sizeof(uint32_t) * nfrom,
+						sizeof(uint32_t) * span,
+						copy_size, in_stride,
+						map_stride);
+				i += span + 2;
+			}
+			if (header->ntrailing > 0) {
+				size_t offset = sfd->buffer_size -
+						header->ntrailing;
+				memcpy(sfd->mem_mirror + offset,
+						act_buffer + header->diff_size,
+						header->ntrailing);
+				stride_shifted_copy(mem_local,
+						(act_buffer + header->diff_size) -
+								offset,
+						offset, header->ntrailing,
+						copy_size, in_stride,
+						map_stride);
+			}
 
-		if (sfd->type == FDC_DMABUF && !already_mapped) {
-			sfd->mem_local = NULL;
 			if (unmap_dmabuf(sfd->dmabuf_bo, handle) == -1) {
 				return 0;
 			}
+		} else {
+			DTRACE_PROBE2(waypipe, apply_diff_enter,
+					sfd->buffer_size, header->diff_size);
+			apply_diff(sfd->buffer_size, sfd->mem_mirror,
+					sfd->mem_local, header->diff_size,
+					header->ntrailing, act_buffer);
+			DTRACE_PROBE(waypipe, apply_diff_exit);
 		}
 
 		return 0;
