@@ -195,11 +195,13 @@ static bool needs_vu_flip(uint32_t drm_format)
 
 bool video_supports_dmabuf_format(uint32_t format, uint64_t modifier)
 {
-	if (modifier == DRM_FORMAT_MOD_LINEAR &&
-			drm_to_av(format) != AV_PIX_FMT_NONE) {
-		return true;
+	/* cannot handle CCS modifiers at the moment due to extra 'plane' issues
+	 */
+	if (modifier == fourcc_mod_code(INTEL, 4) /* Y_TILED_CCS */ ||
+			modifier == fourcc_mod_code(INTEL, 5) /* Yf_TILED_CCS */ || modifier == fourcc_mod_code(INTEL, 6) /* Y_TILED_GEN12_RC_CCS */ || modifier == fourcc_mod_code(INTEL, 7) /* Y_TILED_GEN12_MC_CCS */ || modifier == fourcc_mod_code(INTEL, 8) /* Y_TILED_GEN12_RC_CCS_CC */) {
+		return false;
 	}
-	return false;
+	return drm_to_av(format) != AV_PIX_FMT_NONE;
 }
 bool video_supports_shm_format(uint32_t format)
 {
@@ -481,8 +483,8 @@ void destroy_video_data(struct shadow_fd *sfd)
 	}
 }
 
-static void copy_onto_video_mirror(const char *buffer, AVFrame *frame,
-		const struct dmabuf_slice_data *info)
+static void copy_onto_video_mirror(const char *buffer, uint32_t map_stride,
+		AVFrame *frame, const struct dmabuf_slice_data *info)
 {
 	for (int i = 0; i < info->num_planes; i++) {
 		int j = i;
@@ -493,13 +495,16 @@ static void copy_onto_video_mirror(const char *buffer, AVFrame *frame,
 			uint8_t *dst = frame->data[j] +
 				       frame->linesize[j] * (int)r;
 			const char *src = buffer + (size_t)info->offsets[i] +
-					  (size_t)info->strides[i] * r;
-			memcpy(dst, src, (size_t)info->strides[i]);
+					  (size_t)map_stride * r;
+			/* todo: handle multiplanar strides properly */
+			size_t common = (size_t)minu(map_stride,
+					(uint64_t)frame->linesize[j]);
+			memcpy(dst, src, common);
 		}
 	}
 }
-static void copy_from_video_mirror(char *buffer, const AVFrame *frame,
-		const struct dmabuf_slice_data *info)
+static void copy_from_video_mirror(char *buffer, uint32_t map_stride,
+		const AVFrame *frame, const struct dmabuf_slice_data *info)
 {
 	for (int i = 0; i < info->num_planes; i++) {
 		int j = i;
@@ -510,8 +515,11 @@ static void copy_from_video_mirror(char *buffer, const AVFrame *frame,
 			const uint8_t *src = frame->data[j] +
 					     frame->linesize[j] * (int)r;
 			char *dst = buffer + (size_t)info->offsets[i] +
-				    (size_t)info->strides[i] * r;
-			memcpy(dst, src, (size_t)info->strides[i]);
+				    (size_t)map_stride * r;
+			/* todo: handle multiplanar strides properly */
+			size_t common = (size_t)minu(map_stride,
+					(uint64_t)frame->linesize[j]);
+			memcpy(dst, src, common);
 		}
 	}
 }
@@ -1059,12 +1067,13 @@ void collect_video_from_mirror(
 	if (sfd->video_color_context) {
 		/* If using software encoding, need to convert to YUV */
 		void *handle = NULL;
+		uint32_t map_stride = 0;
 		void *data = map_dmabuf(
-				sfd->dmabuf_bo, false, &handle, NULL, NULL);
+				sfd->dmabuf_bo, false, &handle, &map_stride);
 		if (!data) {
 			return;
 		}
-		copy_onto_video_mirror(data, sfd->video_local_frame,
+		copy_onto_video_mirror(data, map_stride, sfd->video_local_frame,
 				&sfd->dmabuf_info);
 		unmap_dmabuf(sfd->dmabuf_bo, handle);
 
@@ -1232,13 +1241,15 @@ void apply_video_packet(struct shadow_fd *sfd, struct render_data *rd,
 				return;
 			}
 			/* Copy data onto DMABUF */
+			uint32_t map_stride = 0;
 			void *handle = NULL;
 			void *data = map_dmabuf(sfd->dmabuf_bo, true, &handle,
-					NULL, NULL);
+					&map_stride);
 			if (!data) {
 				return;
 			}
-			copy_from_video_mirror(data, sfd->video_local_frame,
+			copy_from_video_mirror(data, map_stride,
+					sfd->video_local_frame,
 					&sfd->dmabuf_info);
 			unmap_dmabuf(sfd->dmabuf_bo, handle);
 		} else {
